// implement Tiger_opcodes.v (please read first)

module TigerCPU (
  input clk_i, // clock
  input rst_i, // synchronous reset, clock must be running
  input [11:0] dat_i, // current memory read data
  output call_overflow_o, // too many call levels, halt and light a red LED
  output ret_underflow_o, // too many return levels, halt and light a red LED
  output loadex_o, // a load from memory is executing
  output storex_o, // a store to memory is executing
  output [11:0] adr_o, // current memory address
  output [11:0] dat_o // current memory write data
);

// 4 data stack registers are implemented
// d[0] is the top of stack
reg [11:0] d [0:3];

// 8 program counter stack registers are implemented (7 valid call/ret levels)
// p[0] is the top of stack and active program counter
reg [11:0] p [0:7];

// There are three 4-bit instruction register slots that execute sequentially
// after an instruction packet fetch.
// Slot 1 is the first to execute, and retires by shifting
// slot1 <= slot2 <= slot3 <= NOP.
// A new packet is fetched when the current packet is exhausted or a jump is
// taken.
// If the final instruction in the current packet does not access memory,
// then the next packet will be fetched in parallel with execution of that
// instruction.
// Otherwise a NOP will arrive in slot1, preserving state while the next packet
// is fetched sequentially.
reg [3:0] slot1, slot2, slot3;

// decoded slot1 instructions
`include "Tiger_opcodes.v"
// NOP excluded
wire lit_x = (slot1 == LIT);
wire ld_x = (slot1 == LD);
wire st_x = (slot1 == ST);
wire tt_x = (slot1 == TT);
wire inv_x = (slot1 == INV);
wire inc_x = (slot1 == INC);
wire asr_x = (slot1 == ASR);
wire adx_x = (slot1 == ADX);
wire add_x = (slot1 == ADD);
wire band_x = (slot1 == BAND);
wire bxor_x = (slot1 == BXOR);
wire jz_x = (slot1 == JZ);
wire jmp_x = (slot1 == JMP);
wire call_x = (slot1 == CALL);
wire ret_x = (slot1 == RET);

// control factors
wire call_overflow = call_x & (p[7] != 0); // fatal error
wire ret_underflow = ret_x & (p[1] == 0); // fatal error
wire sane = ~(call_overflow | ret_underflow); // to continue or halt?
wire memop = lit_x | ld_x | st_x; // current instruction accesses memory
wire prefetch = ~memop & (slot2 == NOP) & (slot3 == NOP); // execute||prefetch
wire take_jump = (jz_x & (d[1] == 0)) | jmp_x | call_x | ret_x; // flow control
wire instruction_fetch = take_jump | prefetch; // fetch next instruction packet

// address mux
reg [11:0] adr;
always @* begin
  if (ret_x) adr = p[1];
  else if (ld_x | st_x | take_jump) adr = d[0];
  else adr = p[0];
end

// instruction register
always @(posedge clk_i) begin
  if (rst_i) begin
    slot1 <= NOP;
    slot2 <= NOP;
    slot3 <= NOP;
  end else if (sane) begin
    if (instruction_fetch) begin
      slot1 <= dat_i[11:8];
      slot2 <= dat_i[7:4];
      slot3 <= dat_i[3:0];
    end else begin // retire the current instruction
      slot1 <= slot2;
      slot2 <= slot3;
      slot3 <= NOP;
    end
  end // sane
end // clk

// program counter stack
always @(posedge clk_i) begin
  if (rst_i) begin
    p[0] <= 0; // first instruction packet fetch address
    p[1] <= 0; // invalid return addresses ...
    p[2] <= 0;
    p[3] <= 0;
    p[4] <= 0;
    p[5] <= 0;
    p[6] <= 0;
    p[7] <= 0;
  end else if (sane) begin

    if (instruction_fetch | lit_x) p[0] <= adr + 1;

    if (call_x) begin
      p[7] <= p[6];
      p[6] <= p[5];
      p[5] <= p[4];
      p[4] <= p[3];
      p[3] <= p[2];
      p[2] <= p[1];
      p[1] <= p[0];
    end

    if (ret_x) begin
      p[1] <= p[2];
      p[2] <= p[3];
      p[3] <= p[4];
      p[4] <= p[5];
      p[5] <= p[6];
      p[6] <= p[7];
      p[7] <= 0; // invalid return address
    end
  end // sane
end // clk

// 13-bit adder, bit 12 is carry
wire [12:0] adder_x = {1'b0, d[0]};
wire [12:0] adder_y = {1'b0, d[1]};
wire [12:0] adder_sum = adder_x + adder_y;

// data stack
always @(posedge clk_i) begin
  if (sane) begin

    if (lit_x) begin
      d[3] <= d[2];
      d[2] <= d[1];
      d[1] <= d[0];
      d[0] <= dat_i; 
    end

    if (ld_x) d[0] <= dat_i;

    if (st_x | jz_x) begin
      d[0] <= d[2];
      d[1] <= d[3];
      d[2] <= d[3];
    end

    if (tt_x) begin
      d[3] <= d[2];
      d[2] <= d[1];
      d[1] <= d[0];
    end

    if (inv_x) d[0] <= ~d[0];

    if (inc_x) d[0] <= d[0] + 1;

    if (asr_x) d[0] <= {d[0][11], d[0][11:1]};

    if (adx_x) begin
      d[0] <= {11'b0, adder_sum[12]};
      d[1] <= adder_sum[11:0];
    end

    if (add_x) begin
      d[0] <= adder_sum[11:0];
      d[1] <= d[2];
      d[2] <= d[3];
    end

    if (band_x) begin
      d[0] <= d[0] & d[1];
      d[1] <= d[2];
      d[2] <= d[3];
    end

    if (bxor_x) begin
      d[0] <= d[0] ^ d[1];
      d[1] <= d[2];
      d[2] <= d[3];
    end

    if (jmp_x | call_x) begin
      d[0] <= d[1];
      d[1] <= d[2];
      d[2] <= d[3];
    end
  end // sane
end // clk

// connect module outputs
assign call_overflow_o = call_overflow;
assign ret_underflow_o = ret_underflow;
assign loadex_o = ld_x; // a load is executing
assign storex_o = st_x; // a store is executing
assign adr_o = adr; // multiplexed per current instruction
assign dat_o = d[1]; // relevant when a store is executing

endmodule

