// Tiger is a minimal 12-bit Instruction Set Architecture for deployment on the Lattice iCEstick.
// It's mission is to provide a gratifying assembly language programming user experience:
//   simple to learn
//   simple to extend
//   self-contained utility, no need to snivel on a forum

// The Tiger programming model is:
//   a data stack (d:) of 4 members
//   a program counter stack (p:) of 8 members
//   a physical memory space of 4096 code, data, and io locations
// The top of the program counter stack is the current program counter.
// There is no program status word, but see ADX re: arithmetic carry.

// Code locations contain a packet of three 4-bit instructions.
// During an instruction packet fetch, the current program counter is
// incremented to point to the next instruction packet
// (or the first of any inline data that satisfies instances of the LIT
// instruction in the current packet).

// The behaviour of each instruction is described using the following notation:
//   -> denotes the arguments/results separator
//   addr denotes a pointer to code/data/io
//   d: denotes that the data stack is specified
//     w x y z denote irrelevant arguments, and any that survive in the results
//   p: denotes that the program counter stack is specified
//     pc denotes the top of the program counter stack
//     ... denotes return addresses according to the current call depth
// All instruction effects are explicit, otherwise doing nothing.
// All instructions execute in 1 clock cycle, including JZ whether the jump is taken or not.

localparam NOP  = 4'h0;
// no operation
// This is necessary for the instruction packet sequencing/prefetch logic!

localparam LIT  = 4'h1;
// d: w x y z -> x y z d // grow 1
// p: ... pc -> ... pc+1
// push d addressed by pc, increment pc
// This is the only way to convey a compile-time expression to a run-time
// constant.

localparam LD   = 4'h2;
// d: x y z addr -> x y z d
// load d addressed by addr

localparam ST   = 4'h3;
// d: y z d addr -> y y y z // shrink 2
// store d addressed by addr

localparam TT   = 4'h4;
// d: x y z d -> y z d d // grow 1
// duplicate d (Top of data stack)

localparam INV  = 4'h5;
// d: x y z d -> x y z ~d
// invert all bits of d

localparam INC  = 4'h6;
// d: x y z d -> x y z d+1
// increment d

localparam ASR  = 4'h7;
// d: x y z d -> x y z d>>1
// arithmetic shift right d, the 2's-complement sign is preserved

localparam ADX  = 4'h8;
// d: y z d1 d2 -> y z d1+d2 carry
// extended add(d1, d2), carry is 0 or 1

localparam ADD  = 4'h9;
// d: y z d1 d2 -> y y z d1+d2 // shrink 1
// add(d1, d2)

localparam BAND = 4'hA;
// d: y z d1 d2 -> y y z d1&d2 // shrink 1
// binary and(d1, d2)
// This is useful for testing which bits are one, or for zeroing bits.

localparam BXOR = 4'hB;
// d: y z d1 d2 -> y y z d1^d2 // shrink 1
// binary exclusive-or(d1, d2)
// This is useful for testing which bits are different, or for inverting bits.

localparam JZ   = 4'hC;
// d: y z flag addr -> y y y z // shrink 2
// p: ... pc -> ... flag?pc:addr+1
// if false flag jump to addr, else continue any trailing instruction slots
// This is the only conditional program flow control instruction.

localparam JMP  = 4'hD;
// d: x y z addr -> x x y z // shrink 1
// p: ... pc -> ... addr+1
// unconditional jump to addr, abandon any trailing instruction slots

localparam CALL = 4'hE;
// d: x y z addr -> x x y z // shrink 1
// p: ... pc -> ... pc addr+1 // grow 1
// unconditional call addr, abandon any trailing instruction slots
// Halt if the program counter stack would have overflowed,
// lighting the dedicated red LED.

localparam RET  = 4'hF;
// p: ... addr pc -> 0 ... addr+1 // shrink 1
// unconditional return from subroutine, abandon any trailing instruction slots
// Halt if the program counter stack would have underflowed,
// lighting the dedicated red LED.

