# Python3 assembler for Tiger ISA
# usage: asm <Desktop/TigerSOC-relative_path/yourprog.asm> e.g. apps/yourapp.asm
import sys # for command line arguments
import math # facilitate generation of lookup tables

srcfile = sys.argv[1] # source filename

# 12-bit fundamentals
ALLBITS = 0xfff
SIGNBIT = 0x800
MAXPOS  = 0x7ff
ram = [0 for i in range(0x1000)] # zeroed RAM image

# return a 12-bit evaluation of expr
def bits12(expr):
    assert expr <= ALLBITS
    assert expr >= -SIGNBIT
    return expr & ALLBITS # suppress sign-extension

# slot 1 opcode images
NOP  = 0x000
LIT  = 0x100
LD   = 0x200
ST   = 0x300
TT   = 0x400
INV  = 0x500
INC  = 0x600
ASR  = 0x700
ADX  = 0x800
ADD  = 0x900
BAND = 0xa00
BXOR = 0xb00
JZ   = 0xc00
JMP  = 0xd00
CALL = 0xe00
RET  = 0xf00

# global state variables
slots  = 0 # number of instructions in current packet of 3
packet = 0 # uncommited image, initial NOPs
pptr   = 0 # packet pointer
lptr   = 1 # literal pointer
passn  = 0 # 0: collect forward references, 1: generate final code

# flush packet to ram
def flush_packet():
    global slots, packet, pptr, lptr
    ram[pptr] = packet
    pptr = lptr
    lptr += 1
    packet = 0
    slots = 0

# align to next packet
def align():
    if slots:
        flush_packet()

# define a unique memory label in pass 0
def lbl(string):
    align()
    if passn == 0:
        exists = string in globals()
        if exists:
            print(string + ' already exists!')
        assert not exists
        exec('global ' + string + '; ' + string + ' = pptr')

# forgive a forward reference in pass 0
def fwd(string):
    if passn == 0:
        return 0
    else:
        return eval(string)

# assemble opcode
def asmop(opcode):
    global slots, packet
    packet |= opcode >> (slots * 4)
    slots += 1
    if slots == 3:
        flush_packet()

def nop():
    asmop(NOP)

def lit(expr):
    global lptr
    ram[lptr] = bits12(expr)
    lptr += 1
    asmop(LIT)

def ld():
    asmop(LD)

def st():
    asmop(ST)

def tt():
    asmop(TT)

def inv():
    asmop(INV)

def inc():
    asmop(INC)

def asr():
    asmop(ASR)

def adx():
    asmop(ADX)

def add():
    asmop(ADD)

def band():
    asmop(BAND)

def bxor():
    asmop(BXOR)

def jz():
    asmop(JZ)

def jmp():
    asmop(JMP)
    align()

def call():
    asmop(CALL)
    align()

def ret():
    asmop(RET)
    align()

# assemble a list of data
def data(numlist):
    global pptr, lptr
    align()
    for i in numlist:
        ram[pptr] = bits12(i)
        pptr += 1
        lptr += 1

# assemble a zero-terminated ASCII string
def ascii(string):
    global pptr, lptr
    align()
    for i in range(len(string)):
        num = ord(string[i])
        assert num < 128
        ram[pptr] = num
        pptr += 1
        lptr += 1
    ram[pptr] = 0
    pptr += 1
    lptr += 1

# include a file
def include(string):
    align()
    with open(string, 'r') as f:
        for line in f:
            exec(line.lstrip()) # prevent unexpected indent errors
    align()

# assemble a file
def asm(string):
    global slots, packet, pptr, lptr, passn
    slots  = 0
    packet = 0
    pptr   = 0
    lptr   = 1
    passn  = 0
    include(string)
    slots  = 0
    packet = 0
    pptr   = 0
    lptr   = 1
    passn  = 1
    include(string)
    print(str(pptr) + ' 12-bit locations assembled')

# display used memory
def dump_ram():
    for i in range(pptr):
        print((format(ram[i], '03X') + ' '), end='')
    print('')

# save entire memory for FPGA RAM initialization
def save_hex(string):
    with open(string, 'w') as f:
        for i in range(4096):
            f.write(format(ram[i], '03X') + '\n')

# save app_start..app_end-1 to output file
def save_app(string):
    tx_pair = bytearray(2) # little-endian data
    with open(string, 'wb') as f:
        # write the header
        tx_pair[0] = app_end & 0x3f
        tx_pair[1] = (app_end >> 6) & 0x3f
        f.write(tx_pair)
        # write the payload
        current_pair = app_start
        while current_pair != app_end:
            tx_pair[0] = ram[current_pair] & 0x3f
            tx_pair[1] = (ram[current_pair] >> 6) & 0x3f
            f.write(tx_pair)
            current_pair += 1

# timing characteristics
fclk       = 12000000 # Hz
tclk       = 1 / fclk # seconds
ktlow      = 2**12 # timer_low factor
kthigh     = 2**24 # timer_high factor

# I/O addresses
pio        = 0xfff
timer_high = 0xffe
timer_low  = 0xffd
rx_status  = 0xffc
tx_status  = 0xffb
uart       = 0xffa

asm(srcfile)
print('You can snoop around if you wish with dump_ram(), dir(), etc.')
print("To save a serial bootloader app, type save_app('apps/yourapp.app')")
print("To save an FPGA RAM image (e.g. the serial bootloader), type save_hex('RAM.hex')")
print('Type exit() when done.')