/*
asynchronous receiver, 8 data bits, no parity
the bit rate is determined below
___     ___ ___ ___ ___ ___ ___ ___ _______
   \___X___X___X___X___X___X___X___X___/
   start 0   1   2   3   4   5   6   7  stop (assumed)
*/

module arx (
  input clk_i,
  input rst_i,
  input rxd_i,       // from the USB/UART converter
  input rst_ready_i, // reset ready_o
  output ready_o,    // rx character ready
  output [7:0] dat_o // rx character, buffered
);
// direct outputs
reg ready_o;
reg [7:0] dat_o;

// synchronized rxd_i
reg rxdm1; // sampled rxd_i
always @(posedge clk_i)
  if (rst_i) rxdm1 <= 1;
  else rxdm1 <= rxd_i;
reg rxdm2; // delayed sample
always @(posedge clk_i)
  if (rst_i) rxdm2 <= 1;
  else rxdm2 <= rxdm1;
wire start_edge = rxdm2 & ~rxdm1; // 1 to 0 rxd transition

// "one-hot" states
reg idle, receiving, done;

// bit rate generator
reg [10:0] brg;
wire next_bit = brg == 1249; // (12000000/9600)-1
wire mid_bit = brg == 625; // 1/2 bit time into current bit
always @(posedge clk_i)
  if (rst_i | ~receiving | next_bit) brg <= 0;
  else brg <= brg + 1;

// lsb-first serial in parallel out, bit 0 is the start bit
reg [8:0] sipo;

// rx state machine
always @(posedge clk_i)

  if (rst_i) begin
    idle <= 1;
    receiving <= 0;
    done <= 0;
    ready_o <= 0;
    sipo <= 9'h1FF;
    dat_o <= 8'h00;
  end else begin

    if (idle & start_edge) begin
      idle <= 0;
      receiving <= 1;
    end

    if (receiving) begin
      if (~sipo[0]) begin // start bit reached
        receiving <= 0;
        done <= 1;
      end else if (mid_bit) sipo <= {rxdm1, sipo[8:1]}; // shift current bit
    end

    if (done) begin
      dat_o <= sipo[8:1];
      sipo <= 9'h1FF;
      ready_o <= 1;
      done <= 0;
      idle <= 1;
    end

    if (~done & rst_ready_i) ready_o <= 0;

  end

endmodule